"""
Direct access to the Pwned Passwords API for checking whether a password is compromised.
"""

import hashlib
import sys

import requests

from . import __version__

API_ENDPOINT = "https://api.pwnedpasswords.com/range/{}"
USER_AGENT = "django-pwned/{} (Python/{} | requests/{})".format(
    __version__, "{}.{}.{}".format(*sys.version_info[:3]), requests.__version__
)


class PwnedRequestError(Exception):
    pass


def _get_pwned(prefix, request_timeout: float) -> dict[str, int]:
    """
    Fetches a dict of all hash suffixes from Pwned Passwords for a given SHA-1 prefix.
    """
    try:
        response = requests.get(
            url=API_ENDPOINT.format(prefix), headers={"User-Agent": USER_AGENT}, timeout=request_timeout
        )
        response.raise_for_status()
    except requests.RequestException as e:
        raise PwnedRequestError("Error fetching data from Pwned Passwords API: %r" % e)

    results = {}
    for line in response.text.splitlines():
        line_suffix, times = line.split(":", 1)
        results[line_suffix] = int(times.replace(",", ""))

    return results


def get_pwned_count(password: str, request_timeout: float) -> int:
    """
    Checks a password against the Pwned Passwords database.

    Returns an integer (count of how many times the password appears in the Pwned data set)
    Raises PwnedRequestError if it couldn't get response from API (timeout, HTTP error code).
    """
    if not isinstance(password, str):
        raise TypeError("Password values to check must be Unicode strings.")
    password_hash = hashlib.sha1(password.encode("utf-8")).hexdigest().upper()
    prefix, suffix = password_hash[:5], password_hash[5:]
    return _get_pwned(prefix, request_timeout).get(suffix, 0)



import pytest
from django.core.exceptions import ValidationError

from django_pwned.exceptions import InvalidArgumentsError
from django_pwned.validators import MinimumUniqueCharactersPasswordValidator


def test_default_arguments():
    validator = MinimumUniqueCharactersPasswordValidator()
    with pytest.raises(ValidationError):
        validator.validate("")
    with pytest.raises(ValidationError):
        validator.validate("a12")
    with pytest.raises(ValidationError):
        validator.validate("a12a12a12a12a12a12a12a12a12a12a12a12a12a12a12a12a12a12a12a12a12")
    validator.validate("1a2b")
    validator.validate("aAbB")


def test_invalid_arguments():
    with pytest.raises(InvalidArgumentsError):
        MinimumUniqueCharactersPasswordValidator(min_unique_characters=1)
    MinimumUniqueCharactersPasswordValidator(min_unique_characters=2)


def test_valid_arguments():
    validator = MinimumUniqueCharactersPasswordValidator(min_unique_characters=8)
    with pytest.raises(ValidationError):
        validator.validate("")
    with pytest.raises(ValidationError):
        validator.validate("1a2b3c4")
    with pytest.raises(ValidationError):
        validator.validate("1a2b3c41a2b3c41a2b3c41a2b3c41a2b3c41a2b3c41a2b3c41a2b3c4")
    validator.validate("1a2b3c4d")
    validator.validate("aAbBcCdD")



import pytest
from django.core.exceptions import ValidationError

from django_pwned.exceptions import InvalidArgumentsError
from django_pwned.validators import GitHubLikePasswordValidator


def test_empty_string():
    validator = GitHubLikePasswordValidator()
    with pytest.raises(ValidationError):
        validator.validate("")

    validator = GitHubLikePasswordValidator(min_length=6, safe_length=7)
    with pytest.raises(ValidationError):
        validator.validate("")


def test_default_arguments():
    validator = GitHubLikePasswordValidator()

    with pytest.raises(ValidationError):
        validator.validate("aaaa777")
    validator.validate("aaaa7777")
    with pytest.raises(ValidationError):
        validator.validate("a" * 8)
    with pytest.raises(ValidationError):
        validator.validate("a" * 14)
    validator.validate("a" * 15)
    validator.validate("a" * 16)


def test_invalid_arguments():
    with pytest.raises(InvalidArgumentsError):
        GitHubLikePasswordValidator(min_length=5, safe_length=15)
    with pytest.raises(InvalidArgumentsError):
        GitHubLikePasswordValidator(min_length=8, safe_length=7)
    with pytest.raises(InvalidArgumentsError):
        GitHubLikePasswordValidator(min_length=8, safe_length=8)
    GitHubLikePasswordValidator(min_length=8, safe_length=9)
    GitHubLikePasswordValidator(min_length=8, safe_length=10)


def test_valid_arguments_1():
    validator = GitHubLikePasswordValidator(min_length=6, safe_length=12)

    with pytest.raises(ValidationError):
        validator.validate("aaa77")
    validator.validate("aaa777")
    with pytest.raises(ValidationError):
        validator.validate("a" * 6)
    with pytest.raises(ValidationError):
        validator.validate("a" * 11)
    validator.validate("a" * 12)
    validator.validate("a" * 13)


def test_valid_arguments_2():
    validator = GitHubLikePasswordValidator(min_length=7, safe_length=8)

    with pytest.raises(ValidationError):
        validator.validate("aaa777")
    validator.validate("aaaa777")
    with pytest.raises(ValidationError):
        validator.validate("a" * 7)
    validator.validate("a" * 8)
    validator.validate("a" * 9)



import re

import pytest
import responses
from django.core.exceptions import ValidationError

from django_pwned.api import API_ENDPOINT
from django_pwned.validators import PwnedPasswordValidator


@responses.activate
def test_django_common_passwords():
    responses.add(responses.GET, url=re.compile(r".*"))
    validator = PwnedPasswordValidator()
    with pytest.raises(ValidationError):
        validator.validate("123456")
    with pytest.raises(ValidationError):
        validator.validate("123456789")
    with pytest.raises(ValidationError):
        validator.validate("qwerty")
    with pytest.raises(ValidationError):
        validator.validate("password")
    with pytest.raises(ValidationError):
        validator.validate("qweasdzxc")
    with pytest.raises(ValidationError):
        validator.validate("startfinding")
    with pytest.raises(ValidationError):
        validator.validate("heyhey1")
    # it should not have called requests.get at all
    assert len(responses.calls) == 0


@responses.activate
def test_pwned_api__leaked_password():
    leaked_password = r"haveibeenpwned"
    leaked_password_hash = "667245565F95194F23408B0EA21A0D02C4EEA81D"
    responses.add(
        responses.GET,
        API_ENDPOINT.format(leaked_password_hash[:5]),
        body="\n".join(
            [
                "0BD86C0E684498894064E4AB86B9420CA0E:763",
                "2019C3022C39C4E5FD5A92ECD102E87476D:3",
                "3C56EAB73498B6A58EF5692C4E4937B4466:81",
                "5565F95194F23408B0EA21A0D02C4EEA81D:1",
                "6E9AC9194DA65040917139BA238B3900354:2,103",
                "BB48AB53E4E454BC487CA6400380C05D41A:5",
                "D55A6ED26C1DE9350D40771822316CC4B29:3",
            ]
        ),
    )
    with pytest.raises(ValidationError):
        PwnedPasswordValidator().validate(leaked_password)
    assert len(responses.calls) == 1


@responses.activate
def test_pwned_api__strong_password():
    strong_password = r"SGz=L.%U\;Os$,k]%U2m"
    strong_password_hash = "761E05BF4161AF6CE0DDA796C063B3B5F0F93A4D"
    responses.add(
        responses.GET,
        API_ENDPOINT.format(strong_password_hash[:5]),
        body="\n".join(
            [
                "1C8ED662FF477F5EFB2F43BB5A772877FEF:1",
                "43DF3CB99A3C26781395C72B543E3D9B77A:1",
                "54B2CEEE3E59AACD9CED9BA6A3DAC33CA62:12",
                "6F97235868B9AB0F74702AF7BB5151DB8BE:8",
                "A61E8BD078A4246979AAE41721B795358D5:80",
                "C47A6ED73AD097E959D9417B5E66E22E8F2:5",
                "E82B1DF7CB04693452281037CEAFB29E037:2",
                "FFF5135901E0131D96F2D5B211ACEA61DAE:1",
            ]
        ),
    )
    PwnedPasswordValidator().validate(strong_password)
    assert len(responses.calls) == 1


@responses.activate
def test_pwned_api__count_threshold():
    leaked_password = r"pass-word"
    leaked_password_hash = "43BEF3EAB34187D71D7E1D9CC307C5E7C07665A8"
    responses.add(
        responses.GET,
        API_ENDPOINT.format(leaked_password_hash[:5]),
        body="\n".join(
            [
                "0BD86C0E684498894064E4AB86B9420CA0E:763",
                "3EAB34187D71D7E1D9CC307C5E7C07665A8:3",
                "3C56EAB73498B6A58EF5692C4E4937B4466:81",
                "5565F95194F23408B0EA21A0D02C4EEA81D:1",
                "6E9AC9194DA65040917139BA238B3900354:2,103",
                "BB48AB53E4E454BC487CA6400380C05D41A:5",
                "D55A6ED26C1DE9350D40771822316CC4B29:3",
            ]
        ),
    )
    with pytest.raises(ValidationError):
        PwnedPasswordValidator().validate(leaked_password)
    with pytest.raises(ValidationError):
        PwnedPasswordValidator(count_threshold=2).validate(leaked_password)
    with pytest.raises(ValidationError):
        PwnedPasswordValidator(count_threshold=3).validate(leaked_password)
    PwnedPasswordValidator(count_threshold=4).validate(leaked_password)
    PwnedPasswordValidator(count_threshold=5).validate(leaked_password)
    assert len(responses.calls) == 5


@responses.activate
def test_pwned_api__connection_error():
    strong_password = r"SGz=L.%U\;Os$,k]%U2m"
    PwnedPasswordValidator().validate(strong_password)
    assert len(responses.calls) == 1


@responses.activate
def test_pwned_api__invalid_input():
    binary_password = b"\x23\x25"
    with pytest.raises(TypeError):
        PwnedPasswordValidator().validate(binary_password)  # noqa
    assert len(responses.calls) == 0



import logging
import string

from django.contrib.auth.password_validation import CommonPasswordValidator, MinimumLengthValidator
from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

from django_pwned.exceptions import InvalidArgumentsError

from . import api

log = logging.getLogger(__name__)
common_password_validator = CommonPasswordValidator()


@deconstructible
class PwnedPasswordValidator:
    """
    Password validator which checks Django's list of common passwords and the Pwned Passwords database.
    """

    def __init__(self, request_timeout: float = 1.5, count_threshold: int = 1):
        self.request_timeout = request_timeout
        self.count_threshold = count_threshold

    def validate(self, password: str, user=None):
        # First, check Django's list of common passwords
        common_password_validator.validate(password, user)

        # If password is not in Django's list, check Pwned API
        try:
            count = api.get_pwned_count(password, self.request_timeout)
        except api.PwnedRequestError as e:
            # Gracefully handle timeouts and HTTP error response codes.
            log.warning("Skipped Pwned Passwords check due to error: %r", e)
            return

        if count >= self.count_threshold:
            raise ValidationError(
                _("Password is in a list of passwords commonly used on other websites."),
                code="password_pwned",
            )

    def get_help_text(self):
        return _("Your password can’t be in the list of commonly used passwords on other websites.")


class GitHubLikePasswordValidator:
    """
    Implements this rule by GitHub:

    Make sure password is at least 15 characters OR at least 8 characters including a number and a lowercase letter.
    """

    def __init__(self, min_length: int = 8, safe_length: int = 15):
        if min_length < 6:
            raise InvalidArgumentsError("min_length must be at least 6")
        if safe_length <= min_length:
            raise InvalidArgumentsError("safe_length must be greater than min_length")
        self.min_length_validator = MinimumLengthValidator(min_length=min_length)
        self.safe_length = safe_length

    def validate(self, password: str, user=None):
        self.min_length_validator.validate(password, user)
        # password is at least "min_length" characters
        if len(password) >= self.safe_length:
            # password is at least "safe_length" characters
            return
        if len(set(password) & set(string.ascii_lowercase)) > 0 and len(set(password) & set(string.digits)) > 0:
            # password includes a number and a lowercase letter
            return
        raise ValidationError(
            _("Passwords shorter than %(safe_length)d characters must include a number and a lowercase letter.")
            % {"safe_length": self.safe_length},
            code="password_github_like_validator",
        )


class MinimumUniqueCharactersPasswordValidator:
    """
    Make sure password contains enough unique characters.
    """

    def __init__(self, min_unique_characters: int = 4):
        if min_unique_characters < 2:
            raise InvalidArgumentsError("min_unique_characters must be at least 2")
        self.min_unique_characters = min_unique_characters

    def validate(self, password: str, user=None):
        if len(set(password)) >= self.min_unique_characters:
            # password has at least "min_unique_characters" unique characters
            return
        raise ValidationError(
            _("Make sure password has at least %(min_unique_characters)d unique characters.")
            % {"min_unique_characters": self.min_unique_characters},
            code="password_min_unique_characters_validator",
        )



import os

from setuptools import find_packages, setup

from django_pwned import __version__

with open(os.path.join(os.path.dirname(__file__), "README.md"), encoding="UTF-8") as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

dev_requirements = ["pre-commit", "pytest", "pytest-cov", "pytest-django", "responses"]

setup(
    name="django-pwned",
    version=__version__,
    description="A Django password validator using the Pwned Passwords API to check for compromised passwords.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Mohammad Javad Naderi",
    url="https://github.com/QueraTeam/django-pwned",
    download_url="https://pypi.org/project/django-pwned/",
    packages=find_packages(".", include=("django_pwned", "django_pwned.*")),
    include_package_data=True,
    install_requires=["Django>=3.2", "requests"],
    extras_require={"dev": dev_requirements},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.0",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Internet :: WWW/HTTP :: Session",
        "Topic :: Security",
    ],
)
